<?php declare(strict_types=1);

namespace massivedynamic\lucidity {

use Exception;
use Throwable;

use massivedynamic\lucidity\TemplateEngine;
use massivedynamic\lucidity\Utils;
use massivedynamic\lucidity\containers\ServiceContainer;
use massivedynamic\lucidity\services\Router;
use massivedynamic\lucidity\services\Service;

abstract class Application {
	protected static ?Application $App = null;

	/** @var array<string, mixed>  */
	protected array            $globals;
	protected TemplateEngine   $templateEngine;
	protected ServiceContainer $serviceContainer;
	protected Router|Service   $router;

	public function __construct(private string $templatesPath, string $partialsPath, string $cachePath) {
		set_exception_handler(fn(Throwable $exception) => $this->handleException($exception));

		if (self::$App instanceof Application)
			throw new Exception("Only one instance of `Application` is possible");
		session_start();
		self::$App              = $this;
		$this->globals          = [
			"errors" => []
		];
		$this->templateEngine   = new TemplateEngine($this, $templatesPath, $partialsPath, $cachePath);
		$this->serviceContainer = new ServiceContainer($this);
		$this->router           = $this->serviceContainer->get(Router::class);
	}

	public function fetchService(string $id) : Service | Application {
		return $this->serviceContainer->get($id);
	}

	public function publishGlobal(string $key, mixed $value) : void {
		if ($key === "errors")
			throw new Exception("The 'errors' global can not be accessed directly. Use `Application::addError()` instead.");
		$this->globals[$key] = $value;
	}

	/**
	 * @param array<string> | string | null $errors
	 */
	public function addError(array | string | null $errors) : void {
		if (!$errors)
			return;
		if (!is_array($errors))
			$errors = [$errors];
		$this->globals["errors"] = array_merge((array)$this->globals["errors"], $errors);
	}

	/**
	 * @return array<string, mixed>
	 */
	public function globals() : array {
		return $this->globals;
	}

	private function handleException(Throwable $exception) : void {
		//Utils::Dump($exception->getTrace());

		self::View("LucidityException", [
			"exception" => $exception,
			"trace" => $exception->getTrace(),
			"exceptionType" => $exception::class,
			"lucidityVersion" => \Composer\InstalledVersions::getVersion("massivedynamic/lucidity"),
			"phpVersion" => phpversion()
		]);
	}

	/**
	 * @param string $template
	 * @param array<string, mixed> $context
	 */
	public static function View(string $template, array $context = []) : void {
		self::$App->templateEngine->view($template, $context);
	}

	public static function Instance() : Application {
		return self::$App;
	}
}

}

namespace {

/**
 * @param string $template
 * @param array<string, mixed> $context
 */
function view(string $template, array $context = []) : void {
	massivedynamic\lucidity\Application::View($template, $context);
}

}
