<?php declare(strict_types=1);

namespace massivedynamic\lucidity\services {

use Exception;
use ReflectionFunction;
use ReflectionMethod;

use massivedynamic\lucidity\Application;
use massivedynamic\lucidity\Utils;
use massivedynamic\lucidity\RequestType;
use massivedynamic\lucidity\containers\ControllerContainer;
use massivedynamic\lucidity\services\Request;
use massivedynamic\lucidity\services\Service;

class Router implements Service {
	private ControllerContainer $controllerContainer;
	private array               $routes;

	public function __construct(private Application $app, private Request $request) {
		$this->controllerContainer = new ControllerContainer($app);
		$this->routes              = [
			RequestType::GET->value    => [],
			RequestType::POST->value   => [],
			RequestType::PUT->value    => [],
			RequestType::DELETE->value => [],
		];
	}

	private function addCallback(RequestType $type, string $route, callable | array $callback) : void {
		$this->routes[$type->value][$route] = $callback;
	}

	public function get(string $route, callable | array $callback) : void {
		$this->addCallback(RequestType::GET, $route, $callback);
	}

	public function post(string $route, callable | array $callback) : void {
		$this->addCallback(RequestType::POST, $route, $callback);
	}

	public function put(string $route, callable | array $callback) : void {
		$this->addCallback(RequestType::PUT, $route, $callback);
	}

	public function delete(string $route, callable | array $callback) : void {
		$this->addCallback(RequestType::DELETE, $route, $callback);
	}

	public function view(string $route, string $template, array $context = []) {
		$this->addCallback(RequestType::GET, $route, fn() => view($template, $context));
	}

	public function route() : void {
		$matched = false;
		foreach ($this->routes[$this->request->type->value] as $route => $callback) {
			if ($this->request->match($route)) {
				$matched = true;
				$this->call($callback);
			}
		}

		if (!$matched)
			view("404");
	}

	public static function Redirect(string $target) : void {
		header("Location: $target");
	}

	private function call(callable | array $callback) : void {
		$callbackInfo = null;

		if (is_array($callback)) {
			assert(count($callback) === 2, "only callables can be routed");
			[$class, $method] = $callback;
			$callbackInfo     = new ReflectionMethod($class, $method);
			$arguments        = $this->getArgs($callbackInfo);
			$callbackInfo->invokeArgs($this->controllerContainer->get($class), $arguments);
		} else {
			$callbackInfo = new ReflectionFunction($callback);
			$arguments    = $this->getArgs($callbackInfo);
			$callbackInfo->invokeArgs($arguments);
		}
	}

	private function getArgs(ReflectionFunction | ReflectionMethod $function) : array {
		$arguments = [];
		foreach ($function->getParameters() as $parameter) {
			$type = $parameter->getType();
			if ($type === null)
				throw new Exception("constructor arguments need to name a type");
			array_push($arguments, $this->app->fetchService($type->getName()));
		}

		return $arguments;
	}
}

}

namespace {

function redirect(string $target) {
	massivedynamic\lucidity\services\Router::Redirect($target);
}

}
